import jax
import jax.numpy as jnp
from jax import random
from functools import partial
from typing import Tuple, Optional

@partial(jax.jit, static_argnums=(1,))
def generate_permutation(key, size: int = 256) -> jnp.ndarray:
    """Generate a permutation table for Perlin noise."""
    perm = jnp.arange(size, dtype=jnp.int32)
    return random.permutation(key, perm)

@jax.jit
def fade(t: jnp.ndarray) -> jnp.ndarray:
    """Smoothing function for Perlin noise interpolation."""
    return t * t * t * (t * (t * 6 - 15) + 10)

@jax.jit
def grad(hash: jnp.ndarray, x: jnp.ndarray, y: jnp.ndarray) -> jnp.ndarray:
    """Calculate gradient for Perlin noise."""
    h = hash & 7
    u = jnp.where(h < 4, x, y)
    v = jnp.where(h < 4, y, x)
    return jnp.where(h & 1, -u, u) + jnp.where(h & 2, -2.0 * v, 2.0 * v)

@partial(jax.jit, static_argnums=(1, 2))
def perlin_noise_2d(
    key,
    width: int,
    height: int,
    scale: float = 50.0
) -> jnp.ndarray:
    """
    Generate 2D Perlin noise.
    
    Args:
        key: JAX random key
        width: Width of the output noise
        height: Height of the output noise
        scale: Scale of the noise pattern
        
    Returns:
        2D array of Perlin noise values
    """
    # Generate permutation table
    p = generate_permutation(key)
    p = jnp.concatenate([p, p])
    
    # Create coordinate grid
    x = jnp.arange(width, dtype=jnp.float32)
    y = jnp.arange(height, dtype=jnp.float32)
    X, Y = jnp.meshgrid(x, y)
    
    # Scale coordinates
    X = X / scale
    Y = Y / scale
    
    # Integer coordinates
    X_int = jnp.floor(X).astype(jnp.int32) & 255
    Y_int = jnp.floor(Y).astype(jnp.int32) & 255
    
    # Fractional coordinates
    X_frac = X - jnp.floor(X)
    Y_frac = Y - jnp.floor(Y)
    
    # Fade factors
    u = fade(X_frac)
    v = fade(Y_frac)
    
    # Hash coordinates
    A = p[X_int] + Y_int
    AA = p[A]
    AB = p[A + 1]
    B = p[X_int + 1] + Y_int
    BA = p[B]
    BB = p[B + 1]
    
    # Generate gradients
    g1 = grad(AA, X_frac, Y_frac)
    g2 = grad(BA, X_frac - 1, Y_frac)
    g3 = grad(AB, X_frac, Y_frac - 1)
    g4 = grad(BB, X_frac - 1, Y_frac - 1)
    
    # Interpolate
    t1 = g1 + u * (g2 - g1)
    t2 = g3 + u * (g4 - g3)
    return t1 + v * (t2 - t1)

@partial(jax.jit, static_argnums=(1, 2, 3, 4, 5))
def fractal_noise_2d(
    key,
    width: int,
    height: int,
    octaves: int = 6,
    persistence: float = 0.5,
    scale: float = 800.0
) -> jnp.ndarray:
    """
    Generate 2D fractal noise (fBm) using Perlin noise.
    
    Args:
        key: JAX random key
        width: Width of the output noise
        height: Height of the output noise
        octaves: Number of octaves to combine
        persistence: How much each octave contributes
        scale: Initial scale of the noise pattern
        
    Returns:
        2D array of fractal noise values
    """
    result = jnp.zeros((height, width))
    max_amplitude = 0.0
    
    for i in range(octaves):
        key, subkey = random.split(key)
        amplitude = persistence ** i
        frequency = 2 ** i
        
        octave = perlin_noise_2d(
            subkey,
            width,
            height,
            scale / frequency
        )
        
        result += amplitude * octave
        max_amplitude += amplitude
    
    # Normalize
    return result / max_amplitude

# Example usage
@partial(jax.jit, static_argnums=(1, 2, 3))
def generate_noise_texture(
    key,
    width: int,
    height: int,
    noise_type: str = "perlin"
) -> jnp.ndarray:
    """
    Generate a noise texture using either Perlin or fractal noise.
    
    Args:
        key: JAX random key
        width: Width of the texture
        height: Height of the texture
        noise_type: Either "perlin" or "fractal"
        
    Returns:
        2D array of noise values normalized to [0, 1]
    """
    if noise_type == "perlin":
        noise = perlin_noise_2d(key, width, height)
    else:
        noise = fractal_noise_2d(key, width, height)
    
    # Normalize to [0, 1]
    noise = (noise - jnp.min(noise)) / (jnp.max(noise) - jnp.min(noise))
    return noise


@jax.jit
def grad1d(hash: jnp.ndarray, x: jnp.ndarray) -> jnp.ndarray:
    """Calculate gradient for 1D Perlin noise."""
    h = hash & 15
    grad = 1.0 + (h & 7)  # Gradient value 1-8
    return jnp.where(h & 8, -grad, grad) * x

@partial(jax.jit, static_argnums=(1,))
def perlin_noise_1d(
    key,
    length: int,
    scale: float = 50.0
) -> jnp.ndarray:
    """
    Generate 1D Perlin noise, suitable for elevation profiles.
    
    Args:
        key: JAX random key
        length: Length of the output noise array
        scale: Scale of the noise pattern
        
    Returns:
        1D array of Perlin noise values
    """
    # Generate permutation table
    p = generate_permutation(key)
    p = jnp.concatenate([p, p])
    
    # Create coordinate array
    x = jnp.arange(length, dtype=jnp.float32) / scale
    
    # Integer and fractional coordinates
    x0 = jnp.floor(x).astype(jnp.int32) & 255
    x_frac = x - jnp.floor(x)
    
    # Fade factor
    u = fade(x_frac)
    
    # Hash coordinates
    A = p[x0]
    B = p[x0 + 1]
    
    # Generate gradients and interpolate
    g1 = grad1d(A, x_frac)
    g2 = grad1d(B, x_frac - 1)
    
    return g1 + u * (g2 - g1)

@partial(jax.jit, static_argnums=(1, 2, 3))
def fractal_noise_1d(
    key,
    length: int,
    octaves: int = 6,
    persistence: float = 0.5,
    scale: float = 50.0,
    height_scale: float = 1.0
) -> jnp.ndarray:
    """
    Generate 1D fractal noise (fBm) using Perlin noise, optimized for elevation profiles.
    
    Args:
        key: JAX random key
        length: Length of the output noise array
        octaves: Number of octaves to combine
        persistence: How much each octave contributes
        scale: Initial scale of the noise pattern
        height_scale: Scaling factor for the final height values
        
    Returns:
        1D array of fractal noise values
    """
    result = jnp.zeros(length)
    max_amplitude = 0.0
    
    for i in range(octaves):
        key, subkey = random.split(key)
        amplitude = persistence ** i
        frequency = 2 ** i
        
        octave = perlin_noise_1d(
            subkey,
            length,
            scale / frequency
        )
        
        result += amplitude * octave
        max_amplitude += amplitude
    
    # Normalize and apply height scaling
    return (result / max_amplitude) * height_scale

def generate_wind_field(
    key,
    length: int,
    time_steps: int,
    base_wind: float = 0.0,
    wind_variation: float = 5.0,
    spatial_scale: float = 100.0,
    temporal_scale: float = 50.0,
    octaves: int = 4,
    persistence: float = 0.5
) -> jnp.ndarray:
    """
    Generate a 2D wind field that varies smoothly in both space and time.
    
    Args:
        key: JAX random key
        length: Spatial length of the wind field
        time_steps: Number of time steps to generate
        base_wind: Base wind speed (m/s)
        wind_variation: Maximum variation in wind speed (m/s)
        spatial_scale: Scale of spatial variations
        temporal_scale: Scale of temporal variations
        octaves: Number of octaves for fractal noise
        persistence: Persistence for fractal noise
        
    Returns:
        2D array of wind speeds (time_steps × length)
    """
    # Generate 2D noise field
    noise = fractal_noise_2d(
        key,
        length,
        time_steps,
        octaves,
        persistence,
        spatial_scale
    )
    
    # Scale noise to wind speeds
    wind = base_wind + (noise - 0.5) * 2 * wind_variation
    
    # Add temporal coherence by applying smoothing in time dimension
    # This ensures wind changes gradually over time
    temporal_kernel = jnp.exp(-jnp.arange(5)**2 / (2 * (temporal_scale/100)**2))
    temporal_kernel = temporal_kernel / jnp.sum(temporal_kernel)
    wind = jax.scipy.signal.convolve(
        wind,
        temporal_kernel[:, jnp.newaxis],
        mode='same'
    )
    
    return wind

def generate_elevation_profile(
    key,
    length: int,
    base_height: float = 100.0,
    height_variation: float = 50.0,
    octaves: int = 6,
    persistence: float = 0.5,
    scale: float = 50.0
) -> jnp.ndarray:
    """
    Generate a realistic elevation profile using fractal noise.
    
    Args:
        key: JAX random key
        length: Length of the elevation profile
        base_height: Base elevation level
        height_variation: Maximum variation in height
        octaves: Number of octaves for fractal noise
        persistence: Persistence for fractal noise
        scale: Scale of the noise pattern
        
    Returns:
        1D array of elevation values
    """
    # Generate base terrain using fractal noise
    noise = fractal_noise_1d(
        key,
        length,
        octaves,
        persistence,
        scale,
        height_variation
    )
    
    # Add base height and ensure non-negative elevation
    elevation = base_height + noise
    return jnp.maximum(elevation, 0.0)
