import jax
import jax.numpy as jnp
from jax import grad, jit, vmap
from functools import partial
from typing import Optional, Tuple, Union

def generate_permutation(key: jax.random.PRNGKey, size: int = 256) -> jnp.ndarray:
    """Generate a random permutation table."""
    perm = jax.random.permutation(key, size)
    # Double the permutation to avoid need for wrapping
    return jnp.concatenate([perm, perm])

@partial(jit, static_argnums=(1,))
def generate_gradients(key: jax.random.PRNGKey, dim: int, size: int = 256) -> jnp.ndarray:
    """Generate random unit vectors for gradient table."""
    # Generate random vectors
    grads = jax.random.normal(key, (size, dim))
    # Normalize to unit vectors
    grads = grads / jnp.linalg.norm(grads, axis=1, keepdims=True)
    # Double the gradients to avoid need for wrapping
    return jnp.concatenate([grads, grads], axis=0)

@jit
def fade(t: jnp.ndarray) -> jnp.ndarray:
    """Fade function 6t^5 - 15t^4 + 10t^3."""
    return t * t * t * (t * (t * 6 - 15) + 10)

def perlin_noise(coords: jnp.ndarray, 
                 gradients: jnp.ndarray,
                 perm: jnp.ndarray,
                 octaves: int = 1,
                 persistence: float = 0.5,
                 lacunarity: float = 2.0,
                 ridged: bool = False,
                 billow: bool = False) -> jnp.ndarray:
    """
    Generate N-dimensional Perlin noise.
    
    Args:
        coords: Array of shape (..., dim) containing coordinates
        gradients: Gradient vectors of shape (size, dim)
        perm: Permutation table
        octaves: Number of octaves for fractal noise
        persistence: Amplitude multiplier for each octave
        lacunarity: Frequency multiplier for each octave
        ridged: If True, generates ridged multifractal noise
        billow: If True, generates billow noise (squared)
    
    Returns:
        Array of noise values
    """
    def single_noise(p):
        # Get integer coordinates
        pi = jnp.floor(p).astype(jnp.int32)
        
        # Get decimal part
        pf = p - pi
        
        # Generate corner coordinates
        corners = jnp.stack(jnp.meshgrid(
            *[jnp.array([0, 1]) for _ in range(p.shape[0])],
            indexing='ij'
        )).reshape(p.shape[0], -1).T
        
        # Get gradients for each corner
        corner_gradients = jnp.zeros((corners.shape[0], p.shape[0]))
        
        for i in range(corners.shape[0]):
            corner_coords = pi + corners[i]
            # Hash coordinates to get gradient index
            index = corner_coords[0]
            for j in range(1, p.shape[0]):
                index = perm[index] + corner_coords[j]
            index = perm[index] % (gradients.shape[0] // 2)
            corner_gradients = corner_gradients.at[i].set(gradients[index])
        
        # Calculate dot products
        vectors = pf - corners
        dots = jnp.sum(vectors * corner_gradients, axis=1)
        
        # Interpolate
        fade_coords = fade(pf)
        for i in range(p.shape[0]):
            n_corners = 2 ** i
            for j in range(0, dots.shape[0], n_corners * 2):
                t = fade_coords[i]
                dots = dots.at[j:j+n_corners].set(
                    dots[j:j+n_corners] + t * (dots[j+n_corners:j+2*n_corners] - dots[j:j+n_corners])
                )
        
        return dots[0]
    
    # Vectorize over multiple coordinates
    noise_fn = vmap(single_noise)
    
    # Initialize accumulator
    result = jnp.zeros(coords.shape[:-1])
    frequency = 1.0
    amplitude = 1.0
    max_value = 0.0
    
    # Generate fractal noise
    for _ in range(octaves):
        noise = noise_fn(coords * frequency)
        
        if ridged:
            noise = 1.0 - jnp.abs(noise)
        elif billow:
            noise = noise * noise
            
        result += noise * amplitude
        max_value += amplitude
        
        frequency *= lacunarity
        amplitude *= persistence
    
    # Normalize
    result /= max_value
    return result

# Example usage for 1D noise
key = jax.random.PRNGKey(0)
key1, key2 = jax.random.split(key)

# Generate tables
perm = generate_permutation(key1)
gradients = generate_gradients(key2, dim=1)

# Generate 1D coordinates
x = jnp.linspace(0, 10, 1000)
coords = jnp.expand_dims(x, axis=1)

# Generate different types of noise
basic_noise = perlin_noise(coords, gradients, perm)
fractal_noise = perlin_noise(coords, gradients, perm, octaves=6)
ridged_noise = perlin_noise(coords, gradients, perm, octaves=6, ridged=True)
billow_noise = perlin_noise(coords, gradients, perm, octaves=6, billow=True)